/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef AGUIAR_GOPINATH_HH
#define AGUIAR_GOPINATH_HH

#include <string>
#include <algorithm>
#include <cmath>
#include <cstdlib>

#include <gsl/gsl_interp.h>
#include <gsl/gsl_rng.h>

#include "GaussQuadrature.hh"
#include "Interpolator1D.hh"

#define MAX_IT 1000

class AguiarGopinath
{
public:
  enum model_t { Model1, Model2 };
  const model_t model;
  const double gam, r, beta, lambda, delta, mu_g, sigma_z, mu_z, rho_z, sigma_g, rho_g, mu_y;
  double min_a;
  const double max_a, min_y, max_y;
  enum state_t { Repay, Default, WaitRedemption, Redemption };
protected:
  const size_t ngrid_a, ngrid_y, ngrid_a1;
  double *grid_a1, *grid_y;
  const double width;

  Interpolator1D VBint;
  
  GaussQuadrature gauss;

  const double convergence_tol, root_tol;

  bool has_converged_flag;
  
  void quadrature_points_y(double y, double *quad_points) const;

  static const double almost_min_infinity = -1e8;

  struct difference_params
  {
    double a;
    const AguiarGopinath *AG;
  };

  //! Returns VG-VB
  static double difference(double y, void *params);

  struct VG_deriv_params
  {
    double y;
    const AguiarGopinath *AG;
  };

  static double VG_deriv(double a, void *params);

public:
  AguiarGopinath(model_t model_arg, double gam_arg, double beta_arg,
                 double lambda_arg, double delta_arg,
                 double convergence_tol_arg, size_t ngrid_a_arg,
                 size_t ngrid_y_arg);
  AguiarGopinath(const AguiarGopinath &arg);
  virtual ~AguiarGopinath();

  double util(double c) const { return (c <= 0 ? NAN : pow(c, 1.0-gam)/(1.0-gam)); };

  //! Computes GDP
  /*! In model 1, the "y" variable contains "z". In model 2, it contains "log(g)" */
  double gdp(double y) const { return (model == Model1 ? exp(y) : exp(y)/mu_g); };

  virtual double VG(double a, double y) const = 0;
  double VB(double y) const { return VBint.get(y); };
  double V(double a, double y) const { return std::max(VB(y), VG(a,y)); };

  //! Optimal borrowing level (assuming it has a good credit history)
  virtual double a1_dr(double a, double y) const = 0;

  //! Does the country default ? (assuming it has a good credit history)
  bool def(double a, double y) const { return VB(y) > VG(a,y); };

  class UpperBound
  {
  };
  class LowerBound
  {
  };

  //! Returns the GDP at which the country decides to default, given a level of asset
  /*! We need today's output in order to determine the range over which the
      search will be performed */
  double default_threshold(double y, double a1) const throw (UpperBound, LowerBound);

  //! Price function
  double q(double y, double a1, double *deriv = NULL) const;

  //! Annual spread faced by the country at a given borrowing level
  double spread(double y, double a1) const;

  //! Consumption
  double cons(double a, double y, state_t decision) const;

  //! Trade balance over GDP
  double tb_y(double a, double y, state_t decision) const { return 1 - cons(a, y, decision)/gdp(y); }

  //! Generate random simulations of the model
  /*! y, a and d are assumed to have their first value set */
  void simulate(double *y, double *a, state_t *d, size_t n, gsl_rng *rng) const;

  void print_simulation(const double *y, const double *a, const state_t *d, size_t n) const;

  class statistics_t
  {
  public:
    double sigma_y, sigma_c, sigma_TB_y, sigma_Rs;
    double rho_c_y, rho_TB_y_y, rho_Rs_y, rho_Rs_TB_y;
    double E_Rs, defaults_10000, mean_debt_output;
    statistics_t() : sigma_y(0.0), sigma_c(0.0), sigma_TB_y(0.0),
                     sigma_Rs(0.0), rho_c_y(0.0), rho_TB_y_y(0.0),
                     rho_Rs_y(0.0), rho_Rs_TB_y(0.0), E_Rs(0.0),
                     defaults_10000(0.0), mean_debt_output(0.0)
    {
    }
  };

  //! Computes the business cycle statistics for a given simulation
  /*! The values are multiplied by scale and then added the corresonding fields in r */
  void simulation_statistics(const double *y, const double *a, const state_t *d, size_t n, size_t warmup, statistics_t &r, double scale) const;

  //! Computes the average business cycle statistics over a number of simulations
  statistics_t model_statistics(size_t nsims = 500, size_t simlength = 1500, size_t warmup = 1000) const;

  void csv_output(const double *y, const double *a, const state_t *d, size_t n, const std::string &filename) const;

  virtual void compute() = 0;

  bool has_converged() const { return has_converged_flag; };
};

#endif
