/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _COHEN_VILLEMOT_HH
#define _COHEN_VILLEMOT_HH

#include <string>
#include <algorithm>
#include <cmath>
#include <cstdlib>

#include <gsl/gsl_interp.h>
#include <gsl/gsl_rng.h>

#include "GaussQuadrature.hh"
#include "Interpolator2D.hh"

#define MAX_IT 1000

class CohenVillemot
{
public:
  const double gam, r, beta, lambda, delta, mu_g, sigma_g, rho_g, mu_y;
  const double p, q, rho_z, mu_z;
  //! Level of asset after a redemption
  const double recovery;
  const double min_a, max_a, min_y, max_y, min_z, max_z;
  enum state_t { Repay, Default, WaitRedemption, Redemption };
protected:
  const size_t ngrid_a, ngrid_y, ngrid_a1, ngrid_z;
  double *grid_y, *grid_z;
  const double width;

  Interpolator2D VBCint, VBNint;
  
  GaussQuadrature gauss;
  
  const size_t quad_npoints;
  double *quad_weights_normal, *quad_weights_bonus_possible;
  bool *quad_points_bonus_possible_normal, *quad_points_bonus_possible_bonus_possible;
  
  const double convergence_tol, root_tol;

  void quadrature_points_B(double y, double z, double *quad_points_y1) const;

  void quadrature_points_G(double y, double z, bool bonus_possible, double *quad_points_y1, double *quad_points_z1) const;

  static const double almost_min_infinity = -1e8;

  struct difference_params
  {
    double a, z;
    bool bonus_possible;
    const CohenVillemot *CV;
  };

  //! Returns VG-VB
  static double difference(double y, void *params);

  struct VG_deriv_params
  {
    double y, z;
    bool bonus_possible;
    const CohenVillemot *CV;
  };

  static double VG_deriv(double a, void *params);

  const static double epsdiff = 1e-8;

public:
  CohenVillemot(double gam_arg, double beta_arg,
                double lambda_arg, double delta_arg,
                double p_arg, double q_arg, double mu_z_arg,
                double recovery_arg, double convergence_tol_arg,
                double min_a_arg, size_t ngrid_a1_arg);
  CohenVillemot(const CohenVillemot &arg);
  virtual ~CohenVillemot();

  double util(double c) const { return (c <= 0 ? NAN : pow(c, 1.0-gam)/(1.0-gam)); };

  //! Computes GDP
  double gdp(double y, double z) const { return (exp(y)+z)/mu_g; };

  virtual double VGN(double a, double y, double z) const = 0;
  virtual double VGC(double a, double y, double z) const = 0;
  double VBN(double y, double z) const;
  double VBC(double y, double z) const;
  double V(double a, double y, double z, bool bonus_possible) const
  {
    if (bonus_possible)
      return std::max(VBC(y, z), VGC(a,y,z));
    else
      return std::max(VBN(y, z), VGN(a,y,z));
  };

  //! Optimal borrowing level (assuming it has a good credit history)
  virtual double a1_dr(double a, double y, double z, bool bonus_possible) const = 0;

  //! Does the country default ? (assuming it has a good credit history)
  bool def(double a, double y, double z, bool bonus_possible) const
  {
    if (bonus_possible)
      return VBC(y, z) > VGC(a,y,z);
    else
      return VBN(y, z) > VGN(a,y,z);
  };

  //! Returns the probability of a repayment tomorrow
  /*! ...given today's y and tomorrow's z, a and bonus_possible */
  double repayment_probability(double y, double z1, double a1, bool bonus_possible1, double *deriv_a1 = NULL) const;

  //! Price function
  double price(double y, double z, bool bonus_possible, double a1, double *deriv_a1 = NULL) const;

  //! Annual spread faced by the country at a given borrowing level
  double spread(double y, double z, bool bonus_possible, double a1) const;

  //! Annualized probability of a crisis tomorrow, given tomorrow's borrowing level
  double crisis_prob(double y, double z, bool bonus_possible, double a1) const;
  
  //! Consumption
  double cons(double a, double y, double z, bool bonus_possible, state_t decision) const;

  //! Trade balance over GDP
  double tb_y(double a, double y, double z, bool bonus_possible, state_t decision) { return 1 - cons(a, y, z, bonus_possible, decision)/gdp(y, z); }

  //! Generate random simulations of the model
  /*! y, a, z and d are assumed to have their first value set */
  void simulate(double *a, double *y, double *z, state_t *d, bool *bonus_possible, size_t n, gsl_rng *rng) const;

  void print_simulation(const double *a, const double *y, const double *z, const state_t *d, const bool *bonus_possible, size_t n) const;

  class statistics_t
  {
  public:
    double sigma_y, sigma_c, sigma_TB_y, sigma_Rs;
    double rho_c_y, rho_TB_y_y, rho_Rs_y, rho_Rs_TB_y;
    double E_Rs, defaults_10000, mean_debt_output, mean_welfare;
    statistics_t() : sigma_y(0.0), sigma_c(0.0), sigma_TB_y(0.0),
                     sigma_Rs(0.0), rho_c_y(0.0), rho_TB_y_y(0.0),
                     rho_Rs_y(0.0), rho_Rs_TB_y(0.0), E_Rs(0.0),
                     defaults_10000(0.0), mean_debt_output(0.0),
                     mean_welfare(0.0)
    {
    }
  };

  //! Computes the business cycle statistics for a given simulation
  /*! The values are multiplied by scale and then added the corresponding fields in r */
  void simulation_statistics(const double *a, const double *y, const double *z, const state_t *d, const bool *bonus_possible, size_t n, size_t warmup, statistics_t &r, double scale) const;

  //! Computes the average business cycle statistics over a number of simulations
  statistics_t model_statistics(size_t nsims = 500, size_t simlength = 1500, size_t warmup = 1000) const;

  void csv_output(const double *a, const double *y, const double *z, const state_t *d, const bool *bonus_possible, size_t n, const std::string &filename) const;

  class ConvergenceError
  {
  public:
    const std::string msg;
    ConvergenceError(const std::string &msg_arg) : msg(msg_arg) {};
  };

  virtual void compute() throw (ConvergenceError) = 0;
};

#endif
