/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <iomanip>

#include <cassert>

#include <ext/algorithm>

#include <gsl/gsl_deriv.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_permute.h>

#include "CohenVillemot2EGM.hh"

//#define DEBUG

CohenVillemot2EGM::CohenVillemot2EGM(double gam_arg, double beta_arg,
                                     double lambda_arg, double delta_arg, double p_arg,
                                     double q_arg, double mu_z_arg, double recovery_arg,
                                     double convergence_tol_arg, double min_a_arg,
                                     bool debt_ceiling_arg)
  : CohenVillemot(gam_arg, beta_arg, lambda_arg, delta_arg, p_arg, q_arg, mu_z_arg,
                  recovery_arg, convergence_tol_arg, min_a_arg, 10),
    debt_ceiling(debt_ceiling_arg),
    min_a_normal(min_a), min_a_bonus_possible(min_a),
    VGNint(ngrid_a, gsl_interp_cspline,
           ngrid_y, gsl_interp_cspline,
           ngrid_z, gsl_interp_cspline,
           false),
    VGCint(ngrid_a, gsl_interp_cspline,
           ngrid_y, gsl_interp_cspline,
           ngrid_z, gsl_interp_cspline,
           false),
    BEV_normal_int(ngrid_a, gsl_interp_cspline,
                   ngrid_y, gsl_interp_cspline,
                   ngrid_z, gsl_interp_cspline),
    BEV_bonus_possible_int(ngrid_a, gsl_interp_cspline,
                           ngrid_y, gsl_interp_cspline,
                           ngrid_z, gsl_interp_cspline),
    a1_dr_normal_int(ngrid_a, gsl_interp_cspline,
                     ngrid_y, gsl_interp_cspline,
                     ngrid_z, gsl_interp_cspline,
                     false),
    a1_dr_bonus_possible_int(ngrid_a, gsl_interp_cspline,
                             ngrid_y, gsl_interp_cspline,
                             ngrid_z, gsl_interp_cspline,
                             false)
{
  assert(ngrid_a == ngrid_a1);

  grid_a1_normal = new double[ngrid_a1];
  grid_a1_bonus_possible = new double[ngrid_a1];
  for (size_t i = 0; i < ngrid_a1; i++)
    {
      grid_a1_normal[i] = min_a+((double) i)/((double) (ngrid_a1-1))*(max_a-min_a);
      grid_a1_bonus_possible[i] = grid_a1_normal[i];
    }

  grid_a_normal = new double[ngrid_a*ngrid_y*ngrid_z];
  grid_a_bonus_possible = new double[ngrid_a*ngrid_y*ngrid_z];
  for (size_t i = 0; i < ngrid_y*ngrid_z; i++)
    {
      __gnu_cxx::copy_n(grid_a1_normal, ngrid_a, grid_a_normal + i * ngrid_a);
      __gnu_cxx::copy_n(grid_a1_bonus_possible, ngrid_a, grid_a_bonus_possible + i * ngrid_a);
    }

  // Initialize VBC
  VBCint.set_grids(grid_y, grid_z);
  for (size_t iz = 0; iz < ngrid_z; iz++)
    for (size_t iy = 0; iy < ngrid_y; iy++)
      VBCint.point(iy, iz) = util(gdp(grid_y[iy], grid_z[iz])+min_a/init_scale_factor)*init_scale_factor;
  VBCint.init();

  // Initialize VBN
  VBNint.set_grids(grid_y, grid_z);
  for (size_t iz = 0; iz < ngrid_z; iz++)
    for (size_t iy = 0; iy < ngrid_y; iy++)
      VBNint.point(iy, iz) = util(gdp(grid_y[iy], grid_z[iz])+min_a/init_scale_factor)*init_scale_factor;
  VBNint.init();

  // Initialize VGN
  VGNint.set_grids(grid_a_normal, grid_y, grid_z);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t ia = 0; ia < ngrid_a; ia++)
      for (size_t iz = 0; iz < ngrid_z; iz++)
        VGNint.point(ia, iy, iz) = util(gdp(grid_y[iy], grid_z[iz]) + grid_a_normal[ia+ngrid_a*(iy+ngrid_y*iz)]/init_scale_factor)*init_scale_factor;
  VGNint.init();
  
  // Initialize VGC
  VGCint.set_grids(grid_a_bonus_possible, grid_y, grid_z);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t ia = 0; ia < ngrid_a; ia++)
      for (size_t iz = 0; iz < ngrid_z; iz++)
        VGCint.point(ia, iy, iz) = util(gdp(grid_y[iy], grid_z[iz]) + grid_a_bonus_possible[ia+ngrid_a*(iy+ngrid_y*iz)]/init_scale_factor)*init_scale_factor;
  VGCint.init();

  // Initialize BEV
  init_BEV();
}

CohenVillemot2EGM::CohenVillemot2EGM(const CohenVillemot2EGM &arg)
  : CohenVillemot(arg),
    debt_ceiling(arg.debt_ceiling),
    min_a_normal(arg.min_a_normal), min_a_bonus_possible(arg.min_a_bonus_possible),
    VGNint(arg.VGNint),
    VGCint(arg.VGCint),
    BEV_normal_int(arg.BEV_normal_int),
    BEV_bonus_possible_int(arg.BEV_bonus_possible_int),
    a1_dr_normal_int(arg.a1_dr_normal_int),
    a1_dr_bonus_possible_int(arg.a1_dr_bonus_possible_int)
{
  grid_a1_normal = new double[ngrid_a1];
  __gnu_cxx::copy_n(arg.grid_a1_normal, ngrid_a1, grid_a1_normal);
  grid_a1_bonus_possible = new double[ngrid_a1];
  __gnu_cxx::copy_n(arg.grid_a1_bonus_possible, ngrid_a1, grid_a1_bonus_possible);
  grid_a_normal = new double[ngrid_a*ngrid_y*ngrid_z];
  __gnu_cxx::copy_n(arg.grid_a_normal, ngrid_a*ngrid_y*ngrid_z, grid_a_normal);
  grid_a_bonus_possible = new double[ngrid_a*ngrid_y*ngrid_z];
  __gnu_cxx::copy_n(arg.grid_a_bonus_possible, ngrid_a*ngrid_y*ngrid_z, grid_a_bonus_possible);
}

CohenVillemot2EGM::~CohenVillemot2EGM()
{
  delete[] grid_a1_normal;
  delete[] grid_a1_bonus_possible;
  delete[] grid_a_normal;
  delete[] grid_a_bonus_possible;
}

void
CohenVillemot2EGM::init_BEV()
{
  BEV_normal_int.set_grids(grid_a1_normal, grid_y, grid_z);
  BEV_bonus_possible_int.set_grids(grid_a1_bonus_possible, grid_y, grid_z);

  double
    *quad_points_y1_normal = new double[quad_npoints],
    *quad_points_z1_normal = new double[quad_npoints],
    *quad_points_y1_bonus_possible = new double[quad_npoints],
    *quad_points_z1_bonus_possible = new double[quad_npoints];

  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t iz = 0; iz < ngrid_z; iz++)
      {
        double y = grid_y[iy];
        double z = grid_z[iz];
        double g = exp(y)+z;

        quadrature_points_G(y, z, false, quad_points_y1_normal, quad_points_z1_normal);
        quadrature_points_G(y, z, true, quad_points_y1_bonus_possible, quad_points_z1_bonus_possible);

        for (size_t ia = 0; ia < ngrid_a; ia++)
          {
            BEV_normal_int.point(ia, iy, iz) = 0;
            BEV_bonus_possible_int.point(ia, iy, iz) = 0;
            for (size_t i = 0; i < quad_npoints; i++)
              {
                BEV_normal_int.point(ia, iy, iz) += quad_weights_normal[i] * beta * pow(g, 1-gam) * V(grid_a1_normal[ia], quad_points_y1_normal[i], quad_points_z1_normal[i], quad_points_bonus_possible_normal[i]);
                BEV_bonus_possible_int.point(ia, iy, iz) += quad_weights_bonus_possible[i] * beta * pow(g, 1-gam) * V(grid_a1_bonus_possible[ia], quad_points_y1_bonus_possible[i], quad_points_z1_bonus_possible[i], quad_points_bonus_possible_bonus_possible[i]);
              }
        }
    }

  BEV_normal_int.init();
  BEV_bonus_possible_int.init();

  delete[] quad_points_y1_normal;
  delete[] quad_points_z1_normal;
  delete[] quad_points_y1_bonus_possible;
  delete[] quad_points_z1_bonus_possible;
}

double
CohenVillemot2EGM::BEV(double a1, double y, double z, bool bonus_possible) const
{
  return (bonus_possible ? BEV_bonus_possible_int.get(a1, y, z)
          : BEV_normal_int.get(a1, y, z));
}

double
CohenVillemot2EGM::a1_dr(double a, double y, double z, bool bonus_possible) const
{
  return (bonus_possible ? a1_dr_bonus_possible_int.get(a, y, z)
          : a1_dr_normal_int.get(a, y, z));
}

double
CohenVillemot2EGM::BEV_deriv(double a1, void *params)
{
  deriv_params *p = (deriv_params *) params;
  const CohenVillemot2EGM &CV = *p->CV;
  return CV.BEV(a1, p->y, p->z, p->bonus_possible);
}

int
CohenVillemot2EGM::position_a1(double a1, double y, double z, bool bonus_possible) const
{
  double t2 = NAN;
  double t1 = price(y, z, bonus_possible, a1, &t2);

  gsl_function F;
  deriv_params params = { y, z, bonus_possible, this };
  F.params = &params;
  F.function = &CohenVillemot2EGM::BEV_deriv;
  double v2, abserr;
  gsl_deriv_central(&F, a1, epsdiff, &v2, &abserr);

  double g = exp(y)+z;

  double cons = pow(v2/g/(t1+a1*t2), -1/gam);
  
  double a = cons + t1*a1*g - gdp(y, z);

#ifdef DEBUG  
  std::cout << std::fixed << std::setprecision(3)
            << "Test: a1=" << a1 << ": " << (v2 <= 0 || v2/g/(t1+a1*t2) <= 0 || a < (bonus_possible ? min_a_bonus_possible : min_a_normal) ? -1 : (a > max_a ? 1 : 0))
            << " (q=" << t1 << ",q'=" << t2 << ",BEV'=" << v2 << ",c=" << cons
            << ",a=" << a << ")" << std::endl;
#endif
  return (v2 <= 0 || v2/g/(t1+a1*t2) <= 0 || a < (bonus_possible ? min_a_bonus_possible : min_a_normal)
          ? -1 : (a > max_a ? 1 : 0));
}

void
CohenVillemot2EGM::compute() throw (ConvergenceError)
{
  double
    *VBC_points_new = new double[ngrid_y*ngrid_z],
    *VBN_points_new = new double[ngrid_y*ngrid_z],
    *VGN_points_new = new double[ngrid_a*ngrid_y*ngrid_z],
    *VGC_points_new = new double[ngrid_a*ngrid_y*ngrid_z],
    *BEV_normal_points_old = new double[ngrid_a*ngrid_y*ngrid_z],
    *BEV_bonus_possible_points_old = new double[ngrid_a*ngrid_y*ngrid_z];

    // Initialize min_a1 et max_a1
  double *min_a1_normal = new double[ngrid_y*ngrid_z],
    *max_a1_normal = new double[ngrid_y*ngrid_z],
    *min_a1_bonus_possible = new double[ngrid_y*ngrid_z],
    *max_a1_bonus_possible = new double[ngrid_y*ngrid_z];

  size_t *perm = new size_t[ngrid_a];
  
  int it = 0;
  double err = INFINITY;

  while (err > convergence_tol && it < MAX_IT)
    {
      bool convergence_error = false;
#pragma omp parallel shared(VBN_points_new, VBC_points_new, VGN_points_new, VGC_points_new, convergence_error)
      {
        double *quad_points_y1_VB = new double[gauss.npoints];
        
#pragma omp for schedule(dynamic)
        for (size_t iy = 0; iy < ngrid_y; iy++)
          for (size_t iz = 0; iz < ngrid_z; iz++)
            {
              double y = grid_y[iy];
              double z = grid_z[iz];
#ifdef DEBUG
              std::cout << iy+1 << " (y=" << y << "), " << iz+1 << " (z=" << z << ")" << std::endl;
#endif
              double g = exp(y)+z;

              // Compute VBN and VBC
              quadrature_points_B(y, z, quad_points_y1_VB);
              double EVB1 = 0.0, EVG1 = 0.0;
              for (size_t i = 0; i < gauss.npoints; i++)
                {
                  EVB1 += gauss.weight(i)*VBC(quad_points_y1_VB[i], rho_z*z);
                  EVG1 += gauss.weight(i)*VGN(recovery, quad_points_y1_VB[i], rho_z*z);
                }

              VBC_points_new[iy+iz*ngrid_y] = util((1-delta)*gdp(y, z)) + lambda*beta*pow(g, 1-gam)*EVG1
                + (1-lambda)*beta*pow(g, 1-gam)*EVB1;

              EVB1 = 0.0;
              EVG1 = 0.0;
              for (size_t i = 0; i < gauss.npoints; i++)
                {
                  EVB1 += gauss.weight(i)*VBC(quad_points_y1_VB[i], rho_z*z-mu_z);
                  EVG1 += gauss.weight(i)*VGN(recovery, quad_points_y1_VB[i], rho_z*z-mu_z);
                }

              VBN_points_new[iy+iz*ngrid_y] = util((1-delta)*gdp(y, z)) + lambda*beta*pow(g, 1-gam)*EVG1
                + (1-lambda)*beta*pow(g, 1-gam)*EVB1;

              
              // Compute VGN and VGC
              bool bonus_possible = false;
              do
                {
                  double *min_a1 = bonus_possible ? min_a1_bonus_possible : min_a1_normal,
                    *max_a1 = bonus_possible ? max_a1_bonus_possible : max_a1_normal,
                    *VG_points_new = bonus_possible ? VGC_points_new : VGN_points_new,
                    *grid_a = bonus_possible ? grid_a_bonus_possible : grid_a_normal;
                
                  // Find max.a1
                  double max_a1_sup = max_a, max_a1_inf = bonus_possible ? min_a_bonus_possible : min_a_normal;
                  while (max_a1_sup - max_a1_inf > dichotomy_tol)
                    {
                      double x = (max_a1_sup+max_a1_inf)/2;
                      if (position_a1(x, y, z, bonus_possible) <= 0)
                        max_a1_inf = x;
                      else
                        max_a1_sup = x;
                    }
                  max_a1[iy+ngrid_y*iz] = max_a1_inf;

                  // Find min_a1
                  double min_a1_sup = max_a, min_a1_inf = bonus_possible ? min_a_bonus_possible : min_a_normal;
                  while (min_a1_sup - min_a1_inf > dichotomy_tol)
                    {
                      double x = (min_a1_sup+min_a1_inf)/2;
                      if (position_a1(x, y, z, bonus_possible) >= 0)
                        min_a1_sup = x;
                      else
                        min_a1_inf = x;
                    }
                  min_a1[iy+ngrid_y*iz] = min_a1_sup;
                  
                  if (position_a1(min_a1[iy+ngrid_y*iz], y, z, bonus_possible) != 0
                      || position_a1(max_a1[iy+ngrid_y*iz], y, z, bonus_possible) != 0)
                    {
                      convergence_error = true;
                      goto converr;
                    }
                  assert(min_a1[iy+ngrid_y*iz] < max_a1[iy+ngrid_y*iz]);

                  for (size_t ia = 0; ia < ngrid_a1; ia++)
                    {
                      double a1 = min_a1[iy+ngrid_y*iz]+gsl_pow_3(((double) ia)/((double) (ngrid_a1-1)))*(max_a1[iy+ngrid_y*iz]-min_a1[iy+ngrid_y*iz]);
              
                      double t2 = NAN;
                      double t1 = price(y, z, bonus_possible, a1, &t2);

                      gsl_function F;
                      deriv_params params = { y, z, bonus_possible, this };
                      F.params = &params;
                      F.function = &CohenVillemot2EGM::BEV_deriv;
                      double v2, abserr;
                      gsl_deriv_central(&F, a1, epsdiff, &v2, &abserr);

                      double g = exp(y)+z;

                      double cons = pow(v2/g/(t1+a1*t2), -1/gam);
  
                      //!\\ Global writes
                      grid_a[ia+ngrid_a*(iy+ngrid_y*iz)] = cons + t1*a1*g - gdp(y, z);
                      assert(!std::isnan(grid_a[ia+ngrid_a*(iy+ngrid_y*iz)]));
                      VG_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] = util(cons)+BEV(a1, y, z, bonus_possible);
                      assert(!std::isnan(VG_points_new[ia+ngrid_a*(iy+ngrid_y*iz)]));
#ifdef DEBUG
                      std::cout << std::fixed << std::setprecision(3)
                                << (bonus_possible ? "VGC" : "VGN") << ": a1=" << a1
                                << ", gdp=" << gdp(y, z) << ", q=" << t1
                                << ", q'=" << t2 << ", BEV'=" << v2 << ", c=" << cons
                                << ", a=" << grid_a[ia+ngrid_a*(iy+ngrid_y*iz)]
                                << ", VG=" << VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)]
                                << std::endl;
#endif
                    }
                  
                  bonus_possible = !bonus_possible;

                }
              while(bonus_possible);
              
            converr:
              ;
            }

        delete[] quad_points_y1_VB;
      }
        
      if (convergence_error)
        {
          delete[] VBN_points_new;
          delete[] VBC_points_new;
          delete[] VGN_points_new;
          delete[] VGC_points_new;
          delete[] BEV_normal_points_old;
          delete[] BEV_bonus_possible_points_old;
          delete[] perm;
          delete[] min_a1_normal;
          delete[] max_a1_normal;
          delete[] min_a1_bonus_possible;
          delete[] max_a1_bonus_possible;
                
          throw ConvergenceError("min_a or max_a not within the boundaries");
        }
      

      err = 0;
      for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t iz = 0; iz < ngrid_z; iz++)
          {
            err = std::max(err, fabs(VBC_points_new[iy+ngrid_y*iz]-VBCint.point(iy,iz)));
            err = std::max(err, fabs(VBN_points_new[iy+ngrid_y*iz]-VBNint.point(iy,iz)));
          }
      VBNint.init(VBN_points_new);
      VBCint.init(VBC_points_new);

      // Test that grids are well formed
      for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t iz = 0; iz < ngrid_z; iz++)
          for (size_t ia = 1; ia < ngrid_a; ia++)
            if (grid_a_normal[ia-1+ngrid_a*(iy+ngrid_y*iz)] >
                grid_a_normal[ia+ngrid_a*(iy+ngrid_y*iz)]
                || grid_a_bonus_possible[ia-1+ngrid_a*(iy+ngrid_y*iz)] >
                grid_a_bonus_possible[ia+ngrid_a*(iy+ngrid_y*iz)])
              {
                delete[] VBN_points_new;
                delete[] VBC_points_new;
                delete[] VGN_points_new;
                delete[] VGC_points_new;
                delete[] BEV_normal_points_old;
                delete[] BEV_bonus_possible_points_old;
                delete[] perm;
                delete[] min_a1_normal;
                delete[] max_a1_normal;
                delete[] min_a1_bonus_possible;
                delete[] max_a1_bonus_possible;
                
                throw ConvergenceError("Endogenous grid is not increasing");
              }
      
      VGNint.set_grids(grid_a_normal, grid_y, grid_z);
      VGNint.init(VGN_points_new);

      VGCint.set_grids(grid_a_bonus_possible, grid_y, grid_z);
      VGCint.init(VGC_points_new);

      for (size_t ia = 0; ia < ngrid_a; ia++)
        for (size_t iy = 0; iy < ngrid_y; iy++)
          for (size_t iz = 0; iz < ngrid_z; iz++)
            {
              BEV_normal_points_old[ia+ngrid_a*(iy+ngrid_y*iz)] = BEV_normal_int.get(grid_a1_normal[ia],grid_y[iy],grid_z[iz]);
              BEV_bonus_possible_points_old[ia+ngrid_a*(iy+ngrid_y*iz)] = BEV_bonus_possible_int.get(grid_a1_bonus_possible[ia],grid_y[iy],grid_z[iz]);
            }

      init_BEV();
      
      for (size_t ia = 0; ia < ngrid_a; ia++)
        for (size_t iy = 0; iy < ngrid_y; iy++)
          for (size_t iz = 0; iz < ngrid_z; iz++)
            {
              err = std::max(err, fabs(BEV_normal_points_old[ia+ngrid_a*(iy+ngrid_y*iz)] - BEV_normal_int.point(ia,iy,iz)));
              err = std::max(err, fabs(BEV_bonus_possible_points_old[ia+ngrid_a*(iy+ngrid_y*iz)] - BEV_bonus_possible_int.point(ia,iy,iz)));
            }

      // Update grids for a1
      min_a_normal = INFINITY;
      min_a_bonus_possible = INFINITY;
      for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t iz = 0; iz < ngrid_z; iz++)
          {
            min_a_normal = std::min(min_a_normal, min_a1_normal[iy+ngrid_y*iz]);
            min_a_bonus_possible = std::min(min_a_bonus_possible, min_a1_bonus_possible[iy+ngrid_y*iz]);
          }

      if (debt_ceiling)
        {
          min_a_normal = std::max(min_a_normal - min_a_update_offset, min_a);
          min_a_bonus_possible = std::max(min_a_bonus_possible - min_a_update_offset, min_a);
        }
      else
        {
          min_a_normal -= min_a_update_offset;
          min_a_bonus_possible -= min_a_update_offset;
        }
      
      for (size_t i = 0; i < ngrid_a1; i++)
        {
          grid_a1_normal[i] = min_a_normal+((double) i)/((double) (ngrid_a1-1))*(max_a-min_a_normal);
          grid_a1_bonus_possible[i] = min_a_bonus_possible+((double) i)/((double) (ngrid_a1-1))*(max_a-min_a_bonus_possible);
        }
      
      // Output
      std::cout << "Iteration " << ++it << ": error=" << std::scientific << err << ", min_a_normal=" << min_a_normal << ", min_a_bonus_possible=" << min_a_bonus_possible << std::endl;
    }

  delete[] VBN_points_new;
  delete[] VBC_points_new;
  delete[] VGN_points_new;
  delete[] VGC_points_new;
  delete[] BEV_normal_points_old;
  delete[] BEV_bonus_possible_points_old;
  delete[] perm;
  
  // Initialize a1_dr
  a1_dr_normal_int.set_grids(grid_a_normal, grid_y, grid_z);
  a1_dr_bonus_possible_int.set_grids(grid_a_bonus_possible, grid_y, grid_z);
  for (size_t ia = 0; ia < ngrid_a1; ia++)
    for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t iz = 0; iz < ngrid_z; iz++)
          {
            a1_dr_normal_int.point(ia, iy, iz) = min_a1_normal[iy+ngrid_y*iz]+gsl_pow_3(((double) ia)/((double) (ngrid_a1-1)))*(max_a1_normal[iy+ngrid_y*iz]-min_a1_normal[iy+ngrid_y*iz]);
            a1_dr_bonus_possible_int.point(ia, iy, iz) = min_a1_bonus_possible[iy+ngrid_y*iz]+gsl_pow_3(((double) ia)/((double) (ngrid_a1-1)))*(max_a1_bonus_possible[iy+ngrid_y*iz]-min_a1_bonus_possible[iy+ngrid_y*iz]);
          }
  a1_dr_normal_int.init();
  a1_dr_bonus_possible_int.init();

  delete[] min_a1_normal;
  delete[] max_a1_normal;
  delete[] min_a1_bonus_possible;
  delete[] max_a1_bonus_possible;

  if (it == MAX_IT)
    throw ConvergenceError("Maximum number of iterations reached");
}
