/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef INTERPOLATOR_1D_HH
#define INTERPOLATOR_1D_HH

#include <gsl/gsl_interp.h>

//! Interpolates a function of one variable
/*! 
  Note that this class is able to interpolate outside the grid domain, using
  a first order Taylor-expansion outside the grid.
*/
class Interpolator1D
{
public:
  //! Main constructor
  /*! 
    \param ngrid_arg Size of grid
    \param type_arg Type of interpolation
  */
  Interpolator1D(size_t ngrid_arg, const gsl_interp_type *type_arg);
  //! Copy constructor
  Interpolator1D(const Interpolator1D &arg);
  //! Destructor
  ~Interpolator1D();

  const size_t ngrid;           //!< Size of grid
  const gsl_interp_type *type;  //!< Type of interpolation

  //! Set the interpolation grid
  /*! 
    \param grid_arg Grid points
  */
  void set_grid(const double *grid_arg);

  //! Inits the internal interpolation engine
  /*!
    Has to be called after interpolation values have been modified through points()
  */
  void init();

  //! Set the value of the function at all grid points
  /*! 
    It also initializes the interpolation engine
    \param points_arg Value of the function at grid points
  */
  void init(const double *points_arg);

  //! Get or set the value of the function at a given grid point
  /*! 
    It is necessary to call init() after this function for proper interpolation computations.
    \param i The index of the grid point
    
    \return An l-value pointing to the value at the requested point
  */
  double &point(size_t i);

  //! Compute the interpolated value at an arbitrary point
  /*! 
    \param x The requested point
    
    \return The interpolated value
  */
  double get(double x) const;
private:
  double *grid;                 //!< Grid points
  double *points;               //!< Value at grid points
  gsl_interp *spline;           //!< Interpolation engine
};

#endif
