/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cassert>

#include <gsl/gsl_math.h>

#include <ext/algorithm>

#include "Interpolator2D.hh"

Interpolator2D::Interpolator2D(size_t ngrid1_arg, const gsl_interp_type *type1_arg,
                               size_t ngrid2_arg, const gsl_interp_type *type2_arg,
                               bool tensor_product_grid_arg)
  : ngrid1(ngrid1_arg), type1(type1_arg),
    ngrid2(ngrid2_arg), type2(type2_arg),
    tensor_product_grid(tensor_product_grid_arg)
{
  if (tensor_product_grid)
    grid1 = new double[ngrid1];
  else
    grid1 = new double[ngrid1*ngrid2];
  grid2 = new double[ngrid2];  
  
  points = new double[ngrid1*ngrid2];
  spline = (gsl_interp **) malloc(sizeof(gsl_interp *)*ngrid2);
  shifts = new size_t[ngrid2];
  for (size_t i = 0; i < ngrid2; i++)
    {
      spline[i] = gsl_interp_alloc(type1, ngrid1);
      shifts[i] = 0;
    }
}

Interpolator2D::Interpolator2D(const Interpolator2D &arg)
  : ngrid1(arg.ngrid1), type1(arg.type1),
    ngrid2(arg.ngrid2), type2(arg.type2),
    tensor_product_grid(arg.tensor_product_grid)
{
  if (tensor_product_grid)
    {
      grid1 = new double[ngrid1];
      __gnu_cxx::copy_n(arg.grid1, ngrid1, grid1);
    }
  else
    {
      grid1 = new double[ngrid1*ngrid2];
      __gnu_cxx::copy_n(arg.grid1, ngrid1*ngrid2, grid1);
    }
  
  grid2 = new double[ngrid2];  
  __gnu_cxx::copy_n(arg.grid2, ngrid2, grid2);

  points = new double[ngrid1*ngrid2];
  __gnu_cxx::copy_n(arg.points, ngrid1*ngrid2, points);
  
  spline = (gsl_interp **) malloc(sizeof(gsl_interp *)*ngrid2);
  shifts = new size_t[ngrid2];
  for (size_t i = 0; i < ngrid2; i++)
    {
      shifts[i] = arg.shifts[i];
      spline[i] = gsl_interp_alloc(type1, ngrid1-arg.shifts[i]);
    }

  init();
}

Interpolator2D::~Interpolator2D()
{
  delete[] grid1;
  delete[] grid2;
  
  delete[] points;
  for (size_t i = 0; i < ngrid2; i++)
    gsl_interp_free(spline[i]);
  free(spline);
  delete[] shifts;
}

void
Interpolator2D::set_grids(const double *grid1_arg, const double *grid2_arg)
{
  if (tensor_product_grid)
    __gnu_cxx::copy_n(grid1_arg, ngrid1, grid1);
  else
    __gnu_cxx::copy_n(grid1_arg, ngrid1*ngrid2, grid1);
  __gnu_cxx::copy_n(grid2_arg, ngrid2, grid2);
}

void
Interpolator2D::init()
{
  for (size_t i = 0; i < ngrid2; i++)
    {
      size_t s = ngrid1;
      while (s > 0 && points[i*ngrid1+s-1] > -INFINITY)
        s--;
    
      if (s != shifts[i])
        {
          if (spline[i] != NULL)
            free(spline[i]);
          if (ngrid1 - s > gsl_interp_type_min_size(type1))
            spline[i] = gsl_interp_alloc(type1, ngrid1 - s);
          else if (ngrid1 - s > gsl_interp_type_min_size(gsl_interp_linear))
            spline[i] = gsl_interp_alloc(gsl_interp_linear, ngrid1 - s);
          else
            spline[i] = NULL;
          shifts[i] = s;
        }

      size_t tensor_offset = tensor_product_grid ? 0 : i*ngrid1;
      if (spline[i] != NULL)
        gsl_interp_init(spline[i], grid1 + tensor_offset + s, points + i*ngrid1 + s, ngrid1 - s);
    }
}

void
Interpolator2D::init(const double *points_arg)
{
  __gnu_cxx::copy_n(points_arg, ngrid1*ngrid2, points);
  init();
}

double &
Interpolator2D::point(size_t i1, size_t i2)
{
  return points[i1 + i2*ngrid1];
}

double
Interpolator2D::get(double x, double y) const
{
  assert(!std::isnan(x) && !std::isnan(y));

  double y1 = std::min(std::max(y, grid2[0]), grid2[ngrid2-1]);

  size_t shift_tmp = 0;
  double *points_tmp = new double[ngrid2];
  
  for (size_t i = 0; i < ngrid2; i++)
    {
      size_t tensor_offset = tensor_product_grid ? 0 : i*ngrid1;

      if (spline[i] == NULL || (shifts[i] > 0 && x < grid1[tensor_offset + shifts[i]]))
        shift_tmp = std::max(i+1, shift_tmp);
      else
        {
          double x1 = std::min(std::max(x, grid1[tensor_offset]), grid1[tensor_offset+ngrid1-1]);
          points_tmp[i] = gsl_interp_eval(spline[i], grid1 + tensor_offset + shifts[i], points + i*ngrid1 + shifts[i], x1, NULL);

          if (x1 != x)
            points_tmp[i] += gsl_interp_eval_deriv(spline[i], grid1 + tensor_offset + shifts[i], points + i*ngrid1 + shifts[i], x1, NULL)*(x-x1);
        }
    }

  if (ngrid2-shift_tmp < std::min(gsl_interp_type_min_size(type2),
                                  gsl_interp_type_min_size(gsl_interp_linear))
      || (shift_tmp > 0 && y1 < grid2[shift_tmp]))
    {
      delete[] points_tmp;
      return -INFINITY;
    }
  
  gsl_interp *spline_tmp;
  if (ngrid2-shift_tmp >= gsl_interp_type_min_size(type2))
   spline_tmp = gsl_interp_alloc(type2, ngrid2-shift_tmp);
  else
   spline_tmp = gsl_interp_alloc(gsl_interp_linear, ngrid2-shift_tmp);
  gsl_interp_init(spline_tmp, grid2 + shift_tmp, points_tmp + shift_tmp, ngrid2 - shift_tmp);
  double v = gsl_interp_eval(spline_tmp, grid2 + shift_tmp, points_tmp + shift_tmp, y1, NULL);

  if (y != y1)
    v += gsl_interp_eval_deriv(spline_tmp, grid2 + shift_tmp, points_tmp + shift_tmp, y1, NULL)*(y-y1);
  
  gsl_interp_free(spline_tmp);
  delete[] points_tmp;
  
  return v;
}
