/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef INTERPOLATOR_2D_HH
#define INTERPOLATOR_2D_HH

#include <gsl/gsl_interp.h>

//! Interpolates a function of two variables
/*! 
  Also contains a mechanism to deal with the case where the function is equal
  to minus infinity in some areas of its domain: the interpolation will still
  work (potentially returning minus infinity).

  If the value of the function is minus infinity in some of its grid points,
  the code assumes that the function is increasing in both of its arguments.

  Note that this class is able to interpolate outside the grid domain, using
  a first order Taylor-expansion outside the grid.

  If tensor_product_grid=false, then the class is able to deal with a grid
  which is not a tensor-product grid. The grid is still of size ngrid1*ngrid2,
  but for each element of grid2, there is a different grid over the first
  dimension. grid2 is therefore still a vector of size ngrid2, but grid1 is a
  matrix of size ngrid1*ngrid2.
*/
class Interpolator2D
{
public:
  //! Main constructor
  /*! 
    \param ngrid1_arg Size of grid in first dimension
    \param type1_arg Type of interpolation in first dimension
    \param ngrid2_arg Size of grid in second dimension
    \param type2_arg Type of interpolation is second dimension
    \param tensor_product_grid_arg Whether the grid is tensor product
  */
  Interpolator2D(size_t ngrid1_arg, const gsl_interp_type *type1_arg,
                 size_t ngrid2_arg, const gsl_interp_type *type2_arg,
                 bool tensor_product_grid_arg = true);
  //! Copy constructor
  Interpolator2D(const Interpolator2D &arg);
  //! Destructor
  ~Interpolator2D();

  const size_t ngrid1;          //!< Size of grid in first dimension
  const gsl_interp_type *type1; //!< Type of interpolation in first dimension

  const size_t ngrid2;          //!< Size of grid in second dimension
  const gsl_interp_type *type2; //!< Type of interpolation in second dimension

  const bool tensor_product_grid; //!< Whether the grid is tensor product

  //! Set the interpolation grid
  /*! 
    \param grid1_arg Grid points in first dimension
    \param grid2_arg Grid points in second dimension
  */
  void set_grids(const double *grid1_arg, const double *grid2_arg);

  //! Inits the internal interpolation engine and computes the shifts
  /*!
    Has to be called after interpolation values have been modified through points()
  */
  void init();

  //! Set the value of the function at all grid points
  /*! 
    It also initializes the interpolation engine and computes the shifts
    \param points_arg Value of the function at grid points (first dim has major order)
  */
  void init(const double *points_arg);

  //! Get or set the value of the function at a given grid point
  /*! 
    It is necessary to call init() after this function for proper interpolation computations.
    \param i1 The first dimension index of the grid point
    \param i2 The second dimension index of the grid point
    
    \return An l-value pointing to the value at the requested point
  */
  double &point(size_t i1, size_t i2);

  //! Compute the interpolated value at an arbitrary point
  /*! 
    \param x First coordinate of requested point
    \param y Second coordinate of requested point
    
    \return The interpolated value
  */
  double get(double x, double y) const;
private:
  double *grid1,                //!< Grid points in first dimension
    *grid2;                     //!< Grid points in second dimension
  double *points;               //!< Value at grid points (first dim has major order)
  gsl_interp **spline;          //!< Array of interpolation engines

  //! Shift index for dealing with -inf values
  /*! 
    shifts[i2] = minimum { i1 such that point(i1, i2) > -inf }
   */
  size_t *shifts;
};

#endif
