/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <R.h>
#include <Rinternals.h>
#include <omp.h>

// Prevent conflict between an R macro and the AguiarGopinath::cons() method
#undef cons

#include "CohenVillemotVFI.hh"
#include "CohenVillemot2EGM.hh"
#include "AguiarGopinathVFI.hh"
#include "AguiarGopinath2EGM.hh"

extern "C" {

  void
  cv_finalizer(SEXP ptr)
  {
    if (TYPEOF(ptr) != EXTPTRSXP)
      {
        Rf_error("Argument should be an external pointer");
        return;
      }

    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(ptr));
    if (!cv)
      return;

    delete cv;

    R_ClearExternalPtr(ptr);
  }
  
  SEXP
  new_cv_model(SEXP gam, SEXP beta, SEXP lambda, SEXP delta,
               SEXP p, SEXP q, SEXP mu_z, SEXP recovery, SEXP convergence_tol,
               SEXP min_a, SEXP debt_ceiling, SEXP sol_method, SEXP num_threads)
  {
    if (TYPEOF(gam) != REALSXP || TYPEOF(beta) != REALSXP || TYPEOF(lambda) != REALSXP
        || TYPEOF(delta) != REALSXP || TYPEOF(p) != REALSXP || TYPEOF(q) != REALSXP
        || TYPEOF(mu_z) != REALSXP || TYPEOF(recovery) != REALSXP
        || TYPEOF(convergence_tol) != REALSXP || TYPEOF(min_a) != REALSXP
        || TYPEOF(debt_ceiling) != LGLSXP
        || TYPEOF(sol_method) != STRSXP || TYPEOF(num_threads) != INTSXP)
      error_return("Arguments must be reals (except the sol.method string, debt.ceiling logical and num.threads integer)");

    if (*INTEGER(num_threads) == 0)
      omp_set_num_threads(omp_get_num_procs());
    else
      omp_set_num_threads(*INTEGER(num_threads));

    CohenVillemot *cv;

    if (STRING_ELT(sol_method, 0) == mkChar("vfi"))
      cv = new CohenVillemotVFI(*REAL(gam), *REAL(beta), *REAL(lambda),
                                *REAL(delta), *REAL(p), *REAL(q),
                                *REAL(mu_z), *REAL(recovery),
                                *REAL(convergence_tol), *REAL(min_a));
    else if (STRING_ELT(sol_method, 0) == mkChar("2egm"))
      cv = new CohenVillemot2EGM(*REAL(gam), *REAL(beta), *REAL(lambda),
                                 *REAL(delta), *REAL(p), *REAL(q),
                                 *REAL(mu_z), *REAL(recovery),
                                 *REAL(convergence_tol), *REAL(min_a), *LOGICAL(debt_ceiling));
    else
      error_return("Unknown solution method (valid values: vfi, 2egm)");

    try
      {
        cv->compute();
      }
    catch (CohenVillemot::ConvergenceError &e)
      {
        delete cv;
        error_return(e.msg.c_str());
      }
        
    SEXP r;
    PROTECT(r = R_MakeExternalPtr(cv, R_NilValue, R_NilValue));
    R_RegisterCFinalizerEx(r, cv_finalizer, TRUE);

    UNPROTECT(1);
    return r;
  }
  
  SEXP
  cv_statistics(SEXP model, SEXP nsims)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    PROTECT(nsims = coerceVector(nsims, INTSXP));
    
    CohenVillemot::statistics_t st = cv->model_statistics(*INTEGER(nsims));

    SEXP r, names;
    PROTECT(r = allocVector(VECSXP, 12));
    PROTECT(names = allocVector(STRSXP, 12));
    SET_VECTOR_ELT(r, 0, ScalarReal(st.sigma_y));
    SET_STRING_ELT(names, 0, mkChar("sigma.y"));
    SET_VECTOR_ELT(r, 1, ScalarReal(st.sigma_c));
    SET_STRING_ELT(names, 1, mkChar("sigma.c"));
    SET_VECTOR_ELT(r, 2, ScalarReal(st.sigma_TB_y));
    SET_STRING_ELT(names, 2, mkChar("sigma.TB.y"));
    SET_VECTOR_ELT(r, 3, ScalarReal(st.sigma_Rs));
    SET_STRING_ELT(names, 3, mkChar("sigma.Rs"));
    SET_VECTOR_ELT(r, 4, ScalarReal(st.rho_c_y));
    SET_STRING_ELT(names, 4, mkChar("rho.c.y"));
    SET_VECTOR_ELT(r, 5, ScalarReal(st.rho_TB_y_y));
    SET_STRING_ELT(names, 5, mkChar("rho.TB.y.y"));
    SET_VECTOR_ELT(r, 6, ScalarReal(st.rho_Rs_y));
    SET_STRING_ELT(names, 6, mkChar("rho.Rs.y"));
    SET_VECTOR_ELT(r, 7, ScalarReal(st.rho_Rs_TB_y));
    SET_STRING_ELT(names, 7, mkChar("rho.Rs.TB.y"));
    SET_VECTOR_ELT(r, 8, ScalarReal(st.E_Rs));
    SET_STRING_ELT(names, 8, mkChar("E.Rs"));
    SET_VECTOR_ELT(r, 9, ScalarReal(st.defaults_10000));
    SET_STRING_ELT(names, 9, mkChar("defaults.10000"));
    SET_VECTOR_ELT(r, 10, ScalarReal(st.mean_debt_output));
    SET_STRING_ELT(names, 10, mkChar("mean.debt.output"));
    SET_VECTOR_ELT(r, 11, ScalarReal(st.mean_welfare));
    SET_STRING_ELT(names, 11, mkChar("mean.welfare"));
    namesgets(r, names);

    UNPROTECT(3);
    
    return r;
  }

  SEXP
  cv_gdp(SEXP model, SEXP y, SEXP z)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));
    
    if (TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP || LENGTH(y) != LENGTH(z))
      error_return("Arguments y and z must be reals and have the same size");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(y)));
    for (R_len_t i = 0; i < LENGTH(y); i++)
      REAL(r)[i] = cv->gdp(REAL(y)[i], REAL(z)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_def(SEXP model, SEXP a, SEXP y, SEXP z, SEXP bonus_possible)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a, y and z must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a)
        || LENGTH(z) != LENGTH(a)
        || LENGTH(bonus_possible) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(LGLSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      LOGICAL(r)[i] = cv->def(REAL(a)[i], REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_a1_dr(SEXP model, SEXP a, SEXP y, SEXP z, SEXP bonus_possible)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a, y and z must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a)
        || LENGTH(z) != LENGTH(a)
        || LENGTH(bonus_possible) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = cv->a1_dr(REAL(a)[i], REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_spread(SEXP model, SEXP y, SEXP z, SEXP bonus_possible, SEXP a1)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a1) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a1, y and z must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a1)
        || LENGTH(z) != LENGTH(a1)
        || LENGTH(bonus_possible) != LENGTH(a1))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a1)));
    for (R_len_t i = 0; i < LENGTH(a1); i++)
      REAL(r)[i] = cv->spread(REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i], REAL(a1)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_crisis_prob(SEXP model, SEXP y, SEXP z, SEXP bonus_possible, SEXP a1)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a1) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a1, y and z must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a1)
        || LENGTH(z) != LENGTH(a1)
        || LENGTH(bonus_possible) != LENGTH(a1))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a1)));
    for (R_len_t i = 0; i < LENGTH(a1); i++)
      REAL(r)[i] = cv->crisis_prob(REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i], REAL(a1)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_get_params(SEXP model)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    SEXP r, names;
    PROTECT(r = allocVector(VECSXP, 14));
    PROTECT(names = allocVector(STRSXP, 14));
    SET_VECTOR_ELT(r, 0, ScalarReal(cv->gam));
    SET_STRING_ELT(names, 0, mkChar("gam"));
    SET_VECTOR_ELT(r, 1, ScalarReal(cv->r));
    SET_STRING_ELT(names, 1, mkChar("r"));
    SET_VECTOR_ELT(r, 2, ScalarReal(cv->beta));
    SET_STRING_ELT(names, 2, mkChar("beta"));
    SET_VECTOR_ELT(r, 3, ScalarReal(cv->lambda));
    SET_STRING_ELT(names, 3, mkChar("lambda"));
    SET_VECTOR_ELT(r, 4, ScalarReal(cv->delta));
    SET_STRING_ELT(names, 4, mkChar("delta"));
    SET_VECTOR_ELT(r, 5, ScalarReal(cv->mu_g));
    SET_STRING_ELT(names, 5, mkChar("mu.g"));
    SET_VECTOR_ELT(r, 6, ScalarReal(cv->sigma_g));
    SET_STRING_ELT(names, 6, mkChar("sigma.g"));
    SET_VECTOR_ELT(r, 7, ScalarReal(cv->rho_g));
    SET_STRING_ELT(names, 7, mkChar("rho.g"));
    SET_VECTOR_ELT(r, 8, ScalarReal(cv->mu_y));
    SET_STRING_ELT(names, 8, mkChar("mu.y"));
    SET_VECTOR_ELT(r, 9, ScalarReal(cv->p));
    SET_STRING_ELT(names, 9, mkChar("p"));
    SET_VECTOR_ELT(r, 10, ScalarReal(cv->q));
    SET_STRING_ELT(names, 10, mkChar("q"));
    SET_VECTOR_ELT(r, 11, ScalarReal(cv->rho_z));
    SET_STRING_ELT(names, 11, mkChar("rho.z"));
    SET_VECTOR_ELT(r, 12, ScalarReal(cv->mu_z));
    SET_STRING_ELT(names, 12, mkChar("mu.z"));
    SET_VECTOR_ELT(r, 13, ScalarReal(cv->recovery));
    SET_STRING_ELT(names, 13, mkChar("recovery"));

    namesgets(r, names);

    UNPROTECT(2);

    return r;
  }

  SEXP
  cv_V(SEXP model, SEXP a, SEXP y, SEXP z, SEXP bonus_possible)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a, y and z must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a)
        || LENGTH(z) != LENGTH(a)
        || LENGTH(bonus_possible) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = cv->V(REAL(a)[i], REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_VGN(SEXP model, SEXP a, SEXP y, SEXP z)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a, y and z must be reals");
    if (LENGTH(y) != LENGTH(a)
        || LENGTH(z) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = cv->VGN(REAL(a)[i], REAL(y)[i], REAL(z)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_VGC(SEXP model, SEXP a, SEXP y, SEXP z)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a, y and z must be reals");
    if (LENGTH(y) != LENGTH(a)
        || LENGTH(z) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = cv->VGC(REAL(a)[i], REAL(y)[i], REAL(z)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_VBC(SEXP model, SEXP y, SEXP z)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments y and z must be reals");
    if (LENGTH(y) != LENGTH(z))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(y)));
    for (R_len_t i = 0; i < LENGTH(y); i++)
      REAL(r)[i] = cv->VBC(REAL(y)[i], REAL(z)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_VBN(SEXP model, SEXP y, SEXP z)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(y) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments y and z must be reals");
    if (LENGTH(y) != LENGTH(z))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(y)));
    for (R_len_t i = 0; i < LENGTH(y); i++)
      REAL(r)[i] = cv->VBN(REAL(y)[i], REAL(z)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_repayment_probability(SEXP model, SEXP y, SEXP z1, SEXP a1, SEXP bonus_possible1, SEXP deriv_a1)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a1) != REALSXP || TYPEOF(z1) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a1, z1 and y must be reals");
    if (TYPEOF(bonus_possible1) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a1)
        || LENGTH(y) != LENGTH(z1)
        || LENGTH(bonus_possible1) != LENGTH(y))
      error_return("Argument vectors must have the same length");

    if (!(TYPEOF(deriv_a1) == NILSXP
          || (TYPEOF(deriv_a1) == REALSXP && LENGTH(deriv_a1) == LENGTH(y))))
      error_return("Argument deriv.a1 must be either NULL or a vector of real of the same length than y");
    
    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a1)));
    for (R_len_t i = 0; i < LENGTH(a1); i++)
      REAL(r)[i] = cv->repayment_probability(REAL(y)[i], REAL(z1)[i], REAL(a1)[i], LOGICAL(bonus_possible1)[i], TYPEOF(deriv_a1) == NILSXP ? NULL : REAL(deriv_a1) + i);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_price(SEXP model, SEXP y, SEXP z, SEXP bonus_possible, SEXP a1, SEXP deriv_a1)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a1) != REALSXP || TYPEOF(z) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a1, z and y must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (LENGTH(y) != LENGTH(a1)
        || LENGTH(y) != LENGTH(z)
        || LENGTH(bonus_possible) != LENGTH(y))
      error_return("Argument vectors must have the same length");

    if (!(TYPEOF(deriv_a1) == NILSXP
          || (TYPEOF(deriv_a1) == REALSXP && LENGTH(deriv_a1) == LENGTH(y))))
      error_return("Argument deriv.a1 must be either NULL or a vector of real of the same length than y");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a1)));
    for (R_len_t i = 0; i < LENGTH(a1); i++)
      REAL(r)[i] = cv->price(REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i], REAL(a1)[i],
                             TYPEOF(deriv_a1) == NILSXP ? NULL : REAL(deriv_a1) + i);
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_cons(SEXP model, SEXP a, SEXP y, SEXP z, SEXP bonus_possible, SEXP d)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(z) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a, z and y must be reals");
    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus_possible must be a logical");
    if (TYPEOF(d) != INTSXP)
      error_return("Argument d must be integer");
    if (LENGTH(y) != LENGTH(a)
        || LENGTH(y) != LENGTH(z)
        || LENGTH(bonus_possible) != LENGTH(y)
        || LENGTH(d) != LENGTH(y))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = cv->cons(REAL(a)[i], REAL(y)[i], REAL(z)[i], LOGICAL(bonus_possible)[i],
                            (CohenVillemot::state_t) (INTEGER(d)[i]));
    UNPROTECT(1);
    return r;
  }

  SEXP
  cv_simulate(SEXP model, SEXP a, SEXP y, SEXP z, SEXP d, SEXP bonus_possible, SEXP seed)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    CohenVillemot *cv = static_cast<CohenVillemot *>(R_ExternalPtrAddr(model));

    if (TYPEOF(y) != REALSXP || TYPEOF(a) != REALSXP || TYPEOF(z) != REALSXP)
      error_return("Arguments a, y and z must be reals");

    if (TYPEOF(d) != INTSXP || TYPEOF(seed) != INTSXP)
      error_return("Arguments d and seed must be integers");

    if (TYPEOF(bonus_possible) != LGLSXP)
      error_return("Argument bonus.possible must be logical");
    
    if (LENGTH(y) != LENGTH(a) || LENGTH(y) != LENGTH(d) || LENGTH(z) != LENGTH(y)
        || LENGTH(bonus_possible) != LENGTH(y))
      error_return("Argument vectors must have the same length");

    gsl_rng *r = gsl_rng_alloc(gsl_rng_default);
    gsl_rng_set(r, *INTEGER(seed));

    bool *bp = new bool[LENGTH(y)];
    for (int i = 0; i < LENGTH(y); i++)
      bp[i] = (bool) LOGICAL(bonus_possible)[i];

    cv->simulate(REAL(a), REAL(y), REAL(z), (CohenVillemot::state_t *) INTEGER(d),
                 bp, LENGTH(y), r);

    for (int i = 0; i < LENGTH(y); i++)
      LOGICAL(bonus_possible)[i] = (int) bp[i];
    
    delete[] bp;

    gsl_rng_free(r);
                 
    return R_NilValue;
  }

  void
  ag_finalizer(SEXP ptr)
  {
    if (TYPEOF(ptr) != EXTPTRSXP)
      {
        Rf_error("Argument should be an external pointer");
        return;
      }

    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(ptr));
    if (!ag)
      return;

    delete ag;

    R_ClearExternalPtr(ptr);
  }

  SEXP
  new_ag_model(SEXP model, SEXP gam, SEXP beta, SEXP lambda, SEXP delta,
               SEXP convergence_tol, SEXP ngrid_a, SEXP ngrid_y, SEXP sol_method,
               SEXP num_threads)
  {
    if (TYPEOF(model) != REALSXP || TYPEOF(gam) != REALSXP || TYPEOF(beta) != REALSXP
        || TYPEOF(lambda) != REALSXP || TYPEOF(delta) != REALSXP
        || TYPEOF(convergence_tol) != REALSXP
        || TYPEOF(ngrid_a) != REALSXP || TYPEOF(ngrid_y) != REALSXP
        || TYPEOF(sol_method) != STRSXP || TYPEOF(num_threads) != INTSXP)
      error_return("Arguments must be reals (except the sol.method string and num.threads integer)");

    if (*INTEGER(num_threads) == 0)
      omp_set_num_threads(omp_get_num_procs());
    else
      omp_set_num_threads(*INTEGER(num_threads));
                        
    AguiarGopinath *ag;
    
    if (STRING_ELT(sol_method, 0) == mkChar("vfi"))
      ag = new AguiarGopinathVFI(*REAL(model) == 1 ? AguiarGopinath::Model1 : AguiarGopinath::Model2,
                                 *REAL(gam), *REAL(beta), *REAL(lambda),
                                 *REAL(delta), *REAL(convergence_tol),
                                 (size_t) *REAL(ngrid_a), (size_t) *REAL(ngrid_y));
    else if (STRING_ELT(sol_method, 0) == mkChar("2egm"))
      ag = new AguiarGopinath2EGM(*REAL(model) == 1 ? AguiarGopinath::Model1 : AguiarGopinath::Model2,
                                  *REAL(gam), *REAL(beta), *REAL(lambda),
                                  *REAL(delta), *REAL(convergence_tol),
                                  (size_t) *REAL(ngrid_a), (size_t) *REAL(ngrid_y));
    else
      error_return("Unknown solution method (valid values: vfi, 2egm)");
    
    ag->compute();

    SEXP r;
    PROTECT(r = R_MakeExternalPtr(ag, R_NilValue, R_NilValue));
    R_RegisterCFinalizerEx(r, ag_finalizer, TRUE);
    
    UNPROTECT(1);
    return r;
  }
  
  SEXP
  ag_statistics(SEXP model, SEXP nsims)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    PROTECT(nsims = coerceVector(nsims, INTSXP));
    
    AguiarGopinath::statistics_t st = ag->model_statistics(*INTEGER(nsims));

    SEXP r, names;
    PROTECT(r = allocVector(VECSXP, 11));
    PROTECT(names = allocVector(STRSXP, 11));
    SET_VECTOR_ELT(r, 0, ScalarReal(st.sigma_y));
    SET_STRING_ELT(names, 0, mkChar("sigma.y"));
    SET_VECTOR_ELT(r, 1, ScalarReal(st.sigma_c));
    SET_STRING_ELT(names, 1, mkChar("sigma.c"));
    SET_VECTOR_ELT(r, 2, ScalarReal(st.sigma_TB_y));
    SET_STRING_ELT(names, 2, mkChar("sigma.TB.y"));
    SET_VECTOR_ELT(r, 3, ScalarReal(st.sigma_Rs));
    SET_STRING_ELT(names, 3, mkChar("sigma.Rs"));
    SET_VECTOR_ELT(r, 4, ScalarReal(st.rho_c_y));
    SET_STRING_ELT(names, 4, mkChar("rho.c.y"));
    SET_VECTOR_ELT(r, 5, ScalarReal(st.rho_TB_y_y));
    SET_STRING_ELT(names, 5, mkChar("rho.TB.y.y"));
    SET_VECTOR_ELT(r, 6, ScalarReal(st.rho_Rs_y));
    SET_STRING_ELT(names, 6, mkChar("rho.Rs.y"));
    SET_VECTOR_ELT(r, 7, ScalarReal(st.rho_Rs_TB_y));
    SET_STRING_ELT(names, 7, mkChar("rho.Rs.TB.y"));
    SET_VECTOR_ELT(r, 8, ScalarReal(st.E_Rs));
    SET_STRING_ELT(names, 8, mkChar("E.Rs"));
    SET_VECTOR_ELT(r, 9, ScalarReal(st.defaults_10000));
    SET_STRING_ELT(names, 9, mkChar("defaults.10000"));
    SET_VECTOR_ELT(r, 10, ScalarReal(st.mean_debt_output));
    SET_STRING_ELT(names, 10, mkChar("mean.debt.output"));
    namesgets(r, names);

    UNPROTECT(3);
    
    return r;
  }

  SEXP
  ag_gdp(SEXP model, SEXP y)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));
    
    if (TYPEOF(y) != REALSXP)
      error_return("Argument y must be real");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(y)));
    for (R_len_t i = 0; i < LENGTH(y); i++)
      REAL(r)[i] = ag->gdp(REAL(y)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  ag_def(SEXP model, SEXP a, SEXP y)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a and y must be reals");
    if (LENGTH(y) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(LGLSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      LOGICAL(r)[i] = ag->def(REAL(a)[i], REAL(y)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  ag_a1_dr(SEXP model, SEXP a, SEXP y)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a and y must be reals");
    if (LENGTH(y) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = ag->a1_dr(REAL(a)[i], REAL(y)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  model_spread(SEXP model, SEXP y, SEXP a1)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a1) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a1 and z must be reals");
    if (LENGTH(y) != LENGTH(a1))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a1)));
    for (R_len_t i = 0; i < LENGTH(a1); i++)
      REAL(r)[i] = ag->spread(REAL(y)[i], REAL(a1)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  ag_get_params(SEXP model)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    SEXP r, names;
    PROTECT(r = allocVector(VECSXP, 9));
    PROTECT(names = allocVector(STRSXP, 9));
    SET_VECTOR_ELT(r, 0, ScalarReal(ag->gam));
    SET_STRING_ELT(names, 0, mkChar("gam"));
    SET_VECTOR_ELT(r, 1, ScalarReal(ag->r));
    SET_STRING_ELT(names, 1, mkChar("r"));
    SET_VECTOR_ELT(r, 2, ScalarReal(ag->beta));
    SET_STRING_ELT(names, 2, mkChar("beta"));
    SET_VECTOR_ELT(r, 3, ScalarReal(ag->lambda));
    SET_STRING_ELT(names, 3, mkChar("lambda"));
    SET_VECTOR_ELT(r, 4, ScalarReal(ag->delta));
    SET_STRING_ELT(names, 4, mkChar("delta"));
    SET_VECTOR_ELT(r, 5, ScalarReal(ag->mu_g));
    SET_STRING_ELT(names, 5, mkChar("mu.g"));
    SET_VECTOR_ELT(r, 6, ScalarReal(ag->sigma_g));
    SET_STRING_ELT(names, 6, mkChar("sigma.g"));
    SET_VECTOR_ELT(r, 7, ScalarReal(ag->rho_g));
    SET_STRING_ELT(names, 7, mkChar("rho.g"));
    SET_VECTOR_ELT(r, 8, ScalarReal(ag->mu_y));
    SET_STRING_ELT(names, 8, mkChar("mu.y"));

    namesgets(r, names);

    UNPROTECT(2);

    return r;
  }

  SEXP
  ag_has_converged(SEXP model)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    return ScalarLogical(ag->has_converged());
  }

  SEXP
  ag_V(SEXP model, SEXP a, SEXP y)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a and y must be reals");
    if (LENGTH(y) != LENGTH(a))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = ag->V(REAL(a)[i], REAL(y)[i]);
    UNPROTECT(1);
    return r;
  }

  SEXP
  ag_q(SEXP model, SEXP y, SEXP a1, SEXP deriv_a1)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a1) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a1 and y must be reals");
    if (LENGTH(y) != LENGTH(a1))
      error_return("Argument vectors must have the same length");

    if (!(TYPEOF(deriv_a1) == NILSXP
          || (TYPEOF(deriv_a1) == REALSXP && LENGTH(deriv_a1) == LENGTH(y))))
      error_return("Argument deriv.a1 must be either NULL or a vector of real of the same length than y");
    
    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a1)));
    for (R_len_t i = 0; i < LENGTH(a1); i++)
      REAL(r)[i] = ag->q(REAL(y)[i], REAL(a1)[i], TYPEOF(deriv_a1) == NILSXP ? NULL : REAL(deriv_a1) + i);
    UNPROTECT(1);
    return r;
  }

  SEXP
  ag_simulate(SEXP model, SEXP y, SEXP a, SEXP d, SEXP seed)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(y) != REALSXP || TYPEOF(a) != REALSXP)
      error_return("Arguments a and y must be reals");

    if (TYPEOF(d) != INTSXP || TYPEOF(seed) != INTSXP)
      error_return("Arguments d and seed must be integers");

    if (LENGTH(y) != LENGTH(a) || LENGTH(y) != LENGTH(d))
      error_return("Argument vectors must have the same length");

    gsl_rng *r = gsl_rng_alloc(gsl_rng_default);
    gsl_rng_set(r, *INTEGER(seed));

    ag->simulate(REAL(y), REAL(a), (AguiarGopinath::state_t *) INTEGER(d), LENGTH(y), r);
    
    gsl_rng_free(r);
                 
    return R_NilValue;
  }

  SEXP
  ag_cons(SEXP model, SEXP a, SEXP y, SEXP d)
  {
    if (TYPEOF(model) != EXTPTRSXP)
      error_return("First argument must be a pointer to a model");
    AguiarGopinath *ag = static_cast<AguiarGopinath *>(R_ExternalPtrAddr(model));

    if (TYPEOF(a) != REALSXP || TYPEOF(y) != REALSXP)
      error_return("Arguments a and y must be reals");
    if (TYPEOF(d) != INTSXP)
      error_return("Argument d must be integer");
    if (LENGTH(y) != LENGTH(a) || LENGTH(y) != LENGTH(d))
      error_return("Argument vectors must have the same length");

    SEXP r;
    PROTECT(r = allocVector(REALSXP, LENGTH(a)));
    for (R_len_t i = 0; i < LENGTH(a); i++)
      REAL(r)[i] = ag->cons(REAL(a)[i], REAL(y)[i], (AguiarGopinath::state_t) (INTEGER(d)[i]));
    UNPROTECT(1);
    return r;
  }
}
