### Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
###
### This program is free software: you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation, either version 3 of the License, or
### (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this program.  If not, see <http://www.gnu.org/licenses/>.

library(numDeriv)
library(statmod)
library(stats)

source("wrapper.R")

check.euler <- function(m, a, y, z, bonus.possible)
  {
    stopifnot(!cv.def(m, a, y, z, bonus.possible))
    
    p <- cv.get.params(m)
    g <- exp(y)+z

    a1 <- cv.a1.dr(m,a,y,z,bonus.possible)
    q.deriv.a1 <- NaN
    q <- cv.price(m, y, z, bonus.possible, a1, q.deriv.a1)
    cons <- cv.gdp(m,y,z)+a-g*a1*q
    lhs <- g*(q.deriv.a1*a1+q)*cons^(-p$gam)

    truncation.window <- 4
    gauss.npoints <- 16
    gauss.points <- gauss.quad(gauss.npoints)$nodes * truncation.window * p$sigma.g
    gauss.weights <- gauss.quad(gauss.npoints)$weights * truncation.window / sqrt(pi) / sqrt(2) * exp(-(truncation.window*gauss.quad(gauss.npoints)$nodes)^2/2)/(1-2*pnorm(-truncation.window*p$sigma.g, sd = p$sigma.g))

    quad.npoints <- 2*gauss.npoints
    quad.points.y <- rep(gauss.points + (1-p$rho.g)*p$mu.y + p$rho.g*y, 2)
    quad.points.bp <- c(rep(TRUE, gauss.npoints), rep(FALSE, gauss.npoints))
    quad.weights <- if (bonus.possible) c(gauss.weights*(1-p$q), gauss.weights*p$q) else c(gauss.weights*p$p, gauss.weights*(1-p$p))
    z.rho.z <- rep(p$rho.z*z, gauss.npoints)
    quad.points.z <- if (bonus.possible) c(z.rho.z, z.rho.z+p$mu.z) else c(z.rho.z-p$mu.z, z.rho.z)
    
    rhs <- 0
    for (i in 1:quad.npoints)
      {
        y1 <- quad.points.y[i]
        z1 <- quad.points.z[i]
        bonus.possible1 <- quad.points.bp[i]
        if (!cv.def(m, a1, y1, z1, bonus.possible1))
          {
            cons1 <- cv.cons(m, a1, y1, z1, bonus.possible1, 0)
            rhs <- rhs + quad.weights[i]*p$beta*g^(1-p$gam)*cons1^(-p$gam)
          }
      }
    
    abs((lhs-rhs)/lhs)
  }

accuracy.checks <- function(model, sim.length = 10000)
  {
    a <- numeric(sim.length)
    a[0] <- 0

    y <- numeric(sim.length)
    y[0] <- 0

    z <- numeric(sim.length)
    z[0] <- 0
    
    d <- integer(sim.length)
    d[0] <- as.integer(0)

    bonus.possible <- logical(sim.length)
    bonus.possible[0] <- FALSE
    
    cv.simulate(model, a, y, z, d, bonus.possible)

    euler.mean <- 0
    euler.max <- -Inf
    count <- 0

    euler <- rep(NA, sim.length)
    
    for (i in 1:(sim.length-1))
      if (d[i+1] == 0 || d[i+1] == 3) # Repay or redemption
        {
          count <- count + 1

          euler[i] <- check.euler(model, a[i], y[i], z[i], bonus.possible[i])
          euler.max <- max(euler.max, euler[i])
          euler.mean <- euler.mean + euler[i]
        }

    euler.mean <- euler.mean / count

    list(euler.mean = log10(euler.mean),
         euler.max = log10(euler.max),
         sim = data.frame(y,z,a,bonus.possible,d,euler))
  }

sol.method <- c("vfi", "2egm")
convergence.tol <- c(2e-2, 1e-3)

sloc.vfi <- system("make cv-vfi-sloc | grep '(SLOC)' | sed 's/^[^[:digit:]]\\+//'", intern = TRUE)
sloc.2egm <- system("make cv-2egm-sloc | grep '(SLOC)' | sed 's/^[^[:digit:]]\\+//'", intern = TRUE)

models <- list()
time.multi <- list()
time.single <- list()
sloc <- list()
moments <- list()
accuracy <- list()

for (i in 1:length(sol.method))
  {
    time.multi[[i]] <- system.time(models[[i]] <- new.cv.model(sol.method = sol.method[i], convergence.tol = convergence.tol[i]))

    time.single[[i]] <- system.time(new.cv.model(sol.method = sol.method[i], convergence.tol = convergence.tol[i], num.threads = 1))
    
    sloc[[i]] <- ifelse(sol.method[i] == "vfi", sloc.vfi, sloc.2egm)
    moments[[i]] <- cv.statistics(models[[i]])
    accuracy[[i]] <- accuracy.checks(models[[i]], sim.length = 10000)
  }

sink("cv-accuracy.tex")
cat("\\begin{tabular}{l", rep("|r", length(sol.method)), "}", sep = "", fill = TRUE)
cat(" &", paste(toupper(sol.method), collapse = " & "), "\\\\ \\hline \\hline", fill = TRUE)
cat("\\textit{Solution characteristics} & & \\\\", fill = TRUE)
###cat("Grid points for $y$ & ", paste(ngrid.y, collapse = " & "), "\\\\", fill = TRUE)
###cat("Grid points for $a$ & ", paste(ngrid.a, collapse = " & "), "\\\\", fill = TRUE)
cat("Convergence criterium (in $\\log_{10}$ units) & ", paste("$", log10(convergence.tol), "$", sep = "", collapse = " & "), "\\\\", fill = TRUE)
cat("Lines of C++ code & ", paste(sloc, collapse = " & "), "\\\\ \\hline", fill = TRUE)
cat("\\textit{Solution time}", rep(" & ", length(sol.method)), "\\\\", fill = TRUE)

time.selector <- function(x) paste(formatC(x[3], digits = 1, format = "f"), "s", sep = "")
cat("Single thread & ", paste(lapply(time.single, time.selector), collapse = " & "), "\\\\", fill = TRUE)
cat("8 threads & ", paste(lapply(time.multi, time.selector), collapse = " & "), "\\\\ \\hline", fill = TRUE)

cat("\\textit{Moments}", rep(" & ", length(sol.method)), "\\\\", fill = TRUE)
cat("Rate of default (\\%, per year) & ", paste("$", lapply(moments, function(x) formatC(x$defaults.10000*4/100, digits = 2, format = "f")), "$", sep = "", collapse = " & "), "\\\\", fill = TRUE)

cat("Mean debt output ratio (\\%, annualized) & ", paste("$", lapply(moments, function(x) formatC(x$mean.debt.output/4*100, digits = 2, format = "f")), "$", sep = "", collapse = " & "), "\\\\", fill = TRUE)

sigma.selector <- function(m) paste("$", lapply(moments, function(x) formatC(x[[m]]*100, digits = 2, format = "f")), "$", sep = "", collapse = " & ")

cat("$\\sigma(y)$ (\\%) & ", sigma.selector("sigma.y"), "\\\\", fill = TRUE)
cat("$\\sigma(c)$ (\\%) & ", sigma.selector("sigma.c"), "\\\\", fill = TRUE)
cat("$\\sigma(TB/y)$ (\\%) & ", sigma.selector("sigma.TB.y"), "\\\\", fill = TRUE)
cat("$\\sigma(R_s)$ (\\%) & ", sigma.selector("sigma.Rs"), "\\\\", fill = TRUE)

rho.selector <- function(m) paste("$", lapply(moments, function(x) formatC(x[[m]], digits = 2, format = "f")), "$", sep = "", collapse = " & ")
cat("$\\rho(c, y)$ & ", rho.selector("rho.c.y"), "\\\\", fill = TRUE)
cat("$\\rho(TB/y, y)$ & ", rho.selector("rho.TB.y.y"), "\\\\", fill = TRUE)
cat("$\\rho(R_s, y)$ & ", rho.selector("rho.Rs.y"), "\\\\", fill = TRUE)
cat("$\\rho(R_s, TB/y)$ & ", rho.selector("rho.Rs.TB.y"), "\\\\ \\hline", fill = TRUE)

cat("\\textit{Euler errors (in $\\log_{10}$ units)}", rep(" & ", length(sol.method)), "\\\\", fill = TRUE)
accuracy.selector <- function(m) paste("$", lapply(accuracy, function(x) formatC(x[[m]], digits = 2, format = "f")), "$", sep = "", collapse = " & ")

cat("Mean & ", accuracy.selector("euler.mean"), "\\\\", fill = TRUE)
cat("Max & ", accuracy.selector("euler.max"), "\\\\ \\hline \\hline", fill = TRUE)
cat("\\end{tabular}")
sink()
